/*
 * vscvs.c:
 * Virtual server for CVS.
 *
 * This is a proxy for cvs pserver which figures out which chroot'd CVS
 * repository to use.
 *
 * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] = "$Id: vscvs.c,v 1.9 2003/12/30 19:12:39 chris Exp $";

#include <sys/types.h>

#include <errno.h>
#include <grp.h>
#include <pwd.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include <sys/fcntl.h>
#include <sys/stat.h>

#define AUTHTIMEOUT 30
#define MAXLINE     1024
#define VCVSTAB     "/etc/vcvstab"

#define CVSBINARY   "/bin/cvs"

enum {auth, verify} request;
char *repos, *username, *password, *repos_owner, *repos_root;
uid_t owner_uid;
gid_t owner_gid;

jmp_buf abortpt;

/* sigalrm_handler:
 * Abort on timeout. */
void sigalrm_handler(int s) {
    siglongjmp(abortpt, 1);
}

/* read_authentication_phase:
 * Read the initial CVS authentication stuff and use it to figure out if we
 * allow the request, and, if we do, which VCVS chroot we should use. Returns
 * 1 on success or 0 if the protocol stuff didn't work out.
 *
 * The CVS authentication protocol sends \n-terminated strings, between
 * markers of the form BEGIN VERIFICATION REQUEST/END VERIFICATION REQUEST or
 * BEGIN AUTH REQUEST/END AUTH REQUEST. Between the markers, repository,
 * username and encoded password are sent. */
int read_authentication_phase(void) {
    char buf[5][MAXLINE];
    int i;

    for (i = 0; i < 5; ++i) {
        int l;
        if (!fgets(buf[i], MAXLINE, stdin)) {
            if (errno)
                syslog(LOG_ERR, "read_authentication_phase: fgets: %m");
            else
                syslog(LOG_ERR, "read_authentication_phase: end of file");
            return 0;
        } else if (buf[i][l = strlen(buf[i]) - 1] != '\n') {
            syslog(LOG_ERR, "read_authentication_phase: overlong line");
            return 0;
        } else buf[i][l] = 0;
    }

    if (strcmp(buf[0], "BEGIN AUTH REQUEST") == 0) {
        if (strcmp(buf[4], "END AUTH REQUEST") != 0) {
            syslog(LOG_ERR, "read_authentication: protocol error: marker line mismatch");
            return 0;
        }
        request = auth;
    } else if (strcmp(buf[0], "BEGIN VERIFICATION REQUEST") == 0) {
        if (strcmp(buf[4], "END VERIFICATION REQUEST") != 0) {
            syslog(LOG_ERR, "read_authentication: protocol error: marker line mismatch");
            return 0;
        }
        request = verify;
    } else {
        syslog(LOG_ERR, "read_authentication: protocol error: unrecognised marker lines");
        return 0;
    }

    /* Obtain data from the authentication request. */
    repos    = strdup(buf[1]);
    username = strdup(buf[2]);
    password = strdup(buf[3]);

    return 1;
}

/* read_vcvstab:
 * Attempt to locate the repository in the vcvstab file. Returns 1 on success
 * or 0 on failure. */
int read_vcvstab(void) {
    char buf[MAXLINE];
    int fd;
    struct stat st;
    FILE *fp;
    int i;
    
    fd = open(VCVSTAB, O_RDONLY);

    /* Open file. */
    if (fd == -1) {
        syslog(LOG_ERR, "read_vcvstab: open(`%s'): %m", VCVSTAB);
        return 0;
    } else if (fstat(fd, &st) == -1) {
        syslog(LOG_ERR, "read_vcvstab: stat(`%s'): %m", VCVSTAB);
        return 0;
    }

    /* Check its permissions. */
    if (st.st_uid != 0 || st.st_gid != 0) {
        syslog(LOG_ERR, "read_vcvstab: %s: owner is %d.%d, should be 0.0", VCVSTAB, (int)st.st_uid, (int)st.st_gid);
        return 0;
    } else if ((st.st_mode & 022) != 0) {
        syslog(LOG_ERR, "read_vcvstab: %s: mode 0%03o is too permissive", VCVSTAB, st.st_mode & 0777);
        return 0;
    }

    /* Associate it with a stdio stream. */
    if (!(fp = fdopen(fd, "rt"))) {
        syslog(LOG_ERR, "read_vcvstab: fdopen: %m");
        return 0;
    }

    i = 1;
    while (fgets(buf, sizeof(buf), fp)) {
        int l;
        char *user, *root;

        if (buf[l = strlen(buf) - 1] != '\n') {
            syslog(LOG_ERR, "read_vcvstab: %s: line %d: overlong line", VCVSTAB, i);
            goto end;
        }
        buf[l] = 0;
            
        user = buf;
        root = strchr(buf, ':');

        if (!root) {
            syslog(LOG_ERR, "read_vcvstab: %s: line %d: no : in line", VCVSTAB, i);
            goto end;
        }
        *(root++) = 0;

        /* Have we identified a possible repository root? */
        l = strlen(root);
        if (strlen(repos) > l && repos[l] == '/' && strncmp(root, repos, strlen(root)) == 0) {
            repos_owner = strdup(user);
            repos_root  = strdup(root);
            syslog(LOG_INFO, "read_vcvstab: repository `%s' corresponds to repos_root `%s' owned by %s", repos, root, user);
            goto end;
        }

        ++i;
    }

end:
    fclose(fp);
    if (repos_owner && repos_root) return 1;
    else return 0;
}

/* repository_sanity_checks:
 * Verify that the repository has sane permissions. */
int repository_sanity_checks(void) {
    struct stat st;

    /* Checks on the repository root. */
    if (stat(repos_root, &st) == -1) {
        syslog(LOG_ERR, "repository_sanity_checks: %s: %m", repos_root);
        return 0;
    } else if (st.st_uid != 0 || st.st_gid != 0) {
        syslog(LOG_ERR, "repository_sanity_checks: %s: owner UID/GID is %d/%d, should be 0/0", repos_root, (int)st.st_uid, (int)st.st_gid);
        return 0;
    } else if ((st.st_mode & 0777) != 0755) {
        syslog(LOG_ERR, "repository_sanity_checks: %s: mode is 0%03o, should be 0755", repos_root, (int)(st.st_mode & 0777));
        return 0;
    }

    /* Checks on the repository itself. */
    if (stat(repos, &st) == -1) {
        syslog(LOG_ERR, "repository_sanity_checks: %s: %m", repos);
        return 0;
    } else if (st.st_uid != owner_uid || st.st_gid != owner_gid) {
        syslog(LOG_ERR, "repository_sanity_checks: %s: owner UID/GID is %d/%d, should be %d/%d", repos_root, (int)st.st_uid, (int)st.st_gid, (int)owner_uid, (int)owner_gid);
        return 0;
    }

    return 1;
}

/* kill_connection:
 * Transmit an appropriate string to the remote end of the link to tell the
 * client to piss off, then quit. */
void kill_connection(void) {
    printf("I HATE YOU\n");
    exit(-1);
}

/* print_pipe:
 * Send a string to a pipe, terminating it with a \n. */
void print_pipe(int fd, char *s) {
    write(fd, s, strlen(s));
    write(fd, "\n", 1);
}

/* do_cvs_proxy:
 * Do the CVS proxying stuff. We proxy stuff from standard input into the
 * child process, since we've already consumed the protocol setup stuff, but
 * we let it write to stdout and stderr as normal. */
void do_cvs_proxy(void) {
    pid_t cvspid;
    int fd_to[2];
    char *argv[] = {CVSBINARY, "--allow-root", repos, "pserver", NULL};
    //char *envp[] = {"PATH=/bin:", "LD_PRELOAD=/lib/useracl.so", NULL, NULL};
    char *envp[] = {"PATH=/bin:", NULL, NULL};
    int i;
    char buf[MAXLINE];
    ssize_t len;
    size_t totalinput = 0;

    if (pipe(fd_to) == -1) {
        fprintf(stderr, "vscvs: pipe: %s\n", strerror(errno));
        syslog(LOG_ERR, "do_cvs_proxy: pipe: %m");
        kill_connection();
    } 

    switch (cvspid = fork()) {
        case -1:
            /* Fork failed. */
            syslog(LOG_ERR, "do_cvs_proxy: fork: %m");
            kill_connection();
            break; /* NOTREACHED */

        case 0:
            /* Child: set up standard input and output. */
            close(fd_to[1]);
            close(0);
            dup2(fd_to[0], 0);

            /* Paranoia. */
            for (i = 3; i < getdtablesize(); ++i)
                close(i);

            /* Actually execute the CVS binary. */
            execve(CVSBINARY, argv, envp);
            _exit(2);

        default:
            syslog(LOG_INFO, "do_cvs_proxy: starting CVS in root %s owned by %s for remote user %s", repos_root, repos_owner, username);

            /* Parent: copy data from stdin to the child. */
            close(fd_to[0]);

            /* Resend the authentication stuff. */
            if (request == auth)
                print_pipe(fd_to[1], "BEGIN AUTH REQUEST");
            else
                print_pipe(fd_to[1], "BEGIN VERIFICATION REQUEST");
            print_pipe(fd_to[1], repos);
            print_pipe(fd_to[1], username);
            print_pipe(fd_to[1], password);
            if (request == auth)
                print_pipe(fd_to[1], "END AUTH REQUEST");
            else
                print_pipe(fd_to[1], "END VERIFICATION REQUEST");

            /* Now proxy everything else through. Note that we break one of
             * the Unix programming commandments here by mixing stdio and
             * unistd I/O. In fact, it should be alright, since after sending
             * the authentication request, the client ought to wait for the `I
             * (LOVE|HATE) YOU' response before sending any more data. Still,
             * this is ugly. */
            do {
                len = read(0, buf, MAXLINE);
                if (len > 0) {
                    if (write(fd_to[1], buf, len) == -1) {
                        syslog(LOG_ERR, "do_cvs_proxy: write: %m");
                        fprintf(stderr, "vscvs: write: %s\n", strerror(errno));
                        break;
                    }
                    totalinput += len;
                }
            } while (len > 0 || (len == -1 && errno == EINTR));

            if (len == -1) {
                syslog(LOG_ERR, "do_cvs_proxy: read: %m");
                fprintf(stderr, "vscvs: read: %s\n", strerror(errno));
            }

            close(fd_to[1]);

            syslog(LOG_INFO, "do_cvs_proxy: finished CVS session in root %s owned by %s for remote user %s; processed %u bytes of input", repos_root, repos_owner, username, (unsigned)totalinput);

	    /* XXX do we want to wait() here? */

            exit(0);
    }
}

/* main:
 * Entry point. */
int main(int argc, char *argv[]) {
    struct passwd *pw;
 
    signal(SIGPIPE, SIG_IGN);

    openlog("vscvs", LOG_NDELAY | LOG_PID, LOG_DAEMON);
    
    /* Set the standard streams to be unbuffered. */
    setvbuf(stdin,  NULL, _IONBF, 0);
    setvbuf(stdout, NULL, _IONBF, 0);
    setvbuf(stderr, NULL, _IONBF, 0);
    
    /* Verify that we have appropriate privileges. */
    if (getuid() != 0 || getgid() != 0) {
        syslog(LOG_ERR, "invoked as UID/GID %d/%d, should be 0/0", (int)getuid(), (int)getgid());
        kill_connection();
    }
    
    /* Read and record the CVS authentication phase. */
    if (sigsetjmp(abortpt, 0)) {
        syslog(LOG_ERR, "timed out authentication request");
        kill_connection();
    }
    signal(SIGALRM, sigalrm_handler);
    alarm(AUTHTIMEOUT);
    if (!read_authentication_phase())
        kill_connection();
    alarm(0);

    /* Try to figure out the repository. */
    if (!read_vcvstab()) {
        syslog(LOG_ERR, "could not find repos_root for repository `%s'", repos);
        kill_connection();
    }

    /* Find out the identity of the user who owns the repository. */
    pw = getpwnam(repos_owner);
    if (!pw) {
        syslog(LOG_ERR, "repos_owner `%s' for repos_root `%s' does not exist", repos_owner, repos_root);
        kill_connection();
    }

    owner_uid = pw->pw_uid;
    owner_gid = pw->pw_gid;

    if (owner_uid == 0 || owner_gid == 0) {
        syslog(LOG_ERR, "repos_owner `%s' has privileged UID/GID %d/%d", repos_owner, (int)owner_uid, (int)owner_gid);
        kill_connection();
    }
 
    /* Do various sanity checks on the repository. */
    if (!repository_sanity_checks())
        kill_connection();

    if (chdir(repos_root) == -1) {
        syslog(LOG_ERR, "chdir(`%s'): %m", repos_root);
        kill_connection();
    }
    
    /* Now do the chroot. */
    if (chroot(repos_root) == -1) {
        syslog(LOG_ERR, "chroot(`%s'): %m", repos_root);
        kill_connection();
    }

    /* Set suitable group membership and drop privileges. */
    if (initgroups(repos_owner, owner_gid) == -1) {
        syslog(LOG_ERR, "set group memberships: %m");
        kill_connection();
    }

    if (setgid(owner_gid) == -1 || setuid(owner_uid) == -1) {
        syslog(LOG_ERR, "drop privileges: %m");
        kill_connection();
    }

    /* Actually start proxying CVS stuff. */
    do_cvs_proxy();

    return 0; /* NOTREACHED */
}
